/*
 * Copyright (c) 2016 BiTel Inc. All rights reserved.
*/
package ru.bitel.mybgbilling.kernel.contract;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.enterprise.context.SessionScoped;
import javax.inject.Named;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.ContractTariff;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.ContractTariffEntry;
import ru.bitel.bgbilling.kernel.contract.api.common.bean.ContractTariffGroup;
import ru.bitel.bgbilling.kernel.contract.api.common.service.ContractTariffService;
import ru.bitel.bgbilling.kernel.directory.api.common.Directory;
import ru.bitel.bgbilling.kernel.tariff.common.bean.TariffGroup;
import ru.bitel.bgbilling.kernel.tariff.common.bean.TariffGroupTariff;
import ru.bitel.bgbilling.kernel.tariff.common.bean.TariffPlan;
import ru.bitel.common.TimeUtils;
import ru.bitel.common.Utils;
import ru.bitel.common.function.Async;
import ru.bitel.common.model.MapHolder;
import ru.bitel.mybgbilling.kernel.common.AbstractBean;
import ru.bitel.mybgbilling.kernel.common.inject.BGInject;
import ru.bitel.mybgbilling.kernel.common.inject.BGInjection;

/**
 * Тарифные планы. Не путайте tariffPlanId и contractTariffId.
 * @author amir
 */
@Named
@SessionScoped
@BGInjection
public class TariffBean
	extends AbstractBean
{
    private static final long serialVersionUID = -772422196781196972L;

    private static final Logger logger = LoggerFactory.getLogger( TariffBean.class );

	/**
	 * Справочник глобальных тарифов.
	 */
	@BGInject(type = TariffPlan.class, module = false)
	private Directory<TariffPlan> tariffPlanDirectory;

	/**
	 * Справочник тарифных групп.
	 */
	@BGInject(type = TariffGroup.class, module = false)
	private Directory<TariffGroup> tariffGroupDirectory;

	/**
	 * Сервис биллинга.
	 */
    @BGInject( module = false )
	private ContractTariffService contractTariffService;

	//////////

	/**
	 * map списка тарифов, на которые можно перейти с тарифа-ключа.<br>
	 * <b>contractTariffId : List&lt;TariffPlan&gt;</b>.
	 */
	private Async<Map<Integer, List<TariffPlan>>> availableTariffPlanMap;

	/**
	 * map списка тарифных групп по тарифу.<br>
	 * <b>tariffPlanId : список групп, в которые он входит</b>.
	 * @see #getTariffGroupByTariff(int)
	 */
	private Async<Map<Integer, List<TariffGroup>>> tariffGroupMap;

	/**
	 * Список тарифов (глобальных и персональных) договора.
	 */
	private Async<List<ContractTariffEntry>> contractTariffEntryList;
	
	/**
	 * Список тарифных групп, подключенных к договору.
	 */
	private Async<List<ContractTariffGroup>> contractTariffGroupList;

	/**
	 * Вызывается автоматически при создании данного Bean.
	 */
	@Override
	protected void init()
		throws BGException
	{
		super.init();

		populate();
	}

	/**
	 * Возвращает глобальный тариф по его ID.
	 * @param id
	 * @return
	 * @throws BGException
	 */
	public TariffPlan getTariffPlan( int id )
		throws BGException
	{
		return tariffPlanDirectory.get( id );
	}

	/**
	 * Возвращает тарифную группу по её ID.
	 * @param id
	 * @return
	 * @throws BGException
	 */
	public TariffGroup getTariffGroup( int id )
		throws BGException
	{
		return tariffGroupDirectory.get( id );
	}

	/**
	 * Возвращает тарифную группу по тарифу с учетом того, какие тарифные группы подключены на договоре.
	 * Если будет указан ID не глобального тарифа, то вернет null.<br>
	 * Используется в {@link TariffChangeBean}.
	 * @param tariffPlanId ID глобального персонального тарифа.
	 * @return
	 * @throws BGException
	 */
    public TariffGroup getTariffGroupByTariff( int tariffPlanId )
        throws BGException
    {
        return Utils.maskNull( tariffGroupMap.get().get( tariffPlanId ) ).stream()
                    .filter( g -> contractTariffGroupList.get().stream()
                                                         .anyMatch( cg -> cg.getTariffGroupId() == g.getId() ) )
                    .findFirst().orElse( null );
    }

	@Deprecated
	public List<TariffPlan> getAvailableTariffPlanList( int contractTariffId )
		throws BGException
	{
		return availableTariffPlanMap.get().get( contractTariffId );
	}

	/**
	 * Возвращает список тарифов, на которые можно перейти с указанного тарифа (с учетом тарифных групп договора).
	 * @param e
	 * @return
	 * @throws BGException
	 */
    public List<TariffPlan> getAvailableTariffPlanList( ContractTariffEntry e )
        throws BGException
    {
        return availableTariffPlanMap.get().get( e.getId() );
    }

    /**
     * Возвращает список тарифов, на которые можно перейти с указанного тарифа (с учетом тарифных групп договора).
     * @param e
     * @return
     * @throws BGException
     */
    public List<TariffPlan> getAvailableTariffPlanList( ContractTariff e )
        throws BGException
    {
        return availableTariffPlanMap.get().get( e.getId() );
    }

    /**
     * Обновление данных этого Bean.
     * @throws BGException
     */
	public void populate()
		throws BGException
	{
		logger.info( "populate" );

		Date now = new Date();

        contractTariffEntryList = Async.of( () -> reverse( contractTariffService.contractTariffEntryList( getContractId(), null, 0, 0 ) ) );

        contractTariffGroupList = Async.of( () -> Utils.maskNull( contractTariffService.contractTariffGroupList( getContractId(), new Date() ) ) );

        availableTariffPlanMap = Async.of( this::computeAvailableTariffPlanMap );
        
        tariffGroupMap = Async.of( () -> computeTariffGroupMap( now ) );
	}

	/**
     * Вычисление map'а списка тарифных планов, на которые можно перейти с contractTariffId-ключа.
     * @param now
     * @return
     * @throws BGException
     * @see {@link #availableTariffPlanMap}
     * @see #getAvailableTariffPlanList(ContractTariffEntry)
     */
    private Map<Integer, List<TariffPlan>> computeAvailableTariffPlanMap()
        throws BGException
    {
        Map<Integer, List<TariffPlan>> availableTariffPlanMap = new HashMap<Integer, List<TariffPlan>>();

        MapHolder<Integer, List<Integer>> mapHolder = contractTariffService.contractTariffChangeAvailableMap( getContractId() );
        Map<Integer, List<Integer>> map = mapHolder.getMap();

        for( Map.Entry<Integer, List<Integer>> e : map.entrySet() )
        {
            List<TariffPlan> list = availableTariffPlanMap.computeIfAbsent( e.getKey(), ( k ) -> new ArrayList<>() );

            for( Integer id : e.getValue() )
            {
                TariffPlan plan = tariffPlanDirectory.get( id );
                if( plan == null )
                {
                    continue;
                }

                list.add( plan );
            }
        }
        
        return availableTariffPlanMap;
    }

	/**
	 * Вычисление map'а списка тарифных групп по тарифу с учетом периода активности привязок.
	 * @param now
	 * @return
	 * @throws BGException
	 * @see {@link #tariffGroupMap}
	 * @see #getTariffGroupByTariff(int)
	 */
    private Map<Integer, List<TariffGroup>> computeTariffGroupMap( final Date now )
        throws BGException
    {
        // tariffPlanId : List<TariffGroup>
        Map<Integer, List<TariffGroup>> tariffGroupMap = new HashMap<Integer, List<TariffGroup>>();

        for( TariffGroup tariffGroup : tariffGroupDirectory.list() )
        {
            for( TariffGroupTariff groupTariff : Utils.maskNull( tariffGroup.getPlanList() ) )
            {
                if( !TimeUtils.dateInRange( now, groupTariff.getDate1(), groupTariff.getDate2() ) )
                {
                    continue;
                }

                tariffGroupMap.computeIfAbsent( groupTariff.getId(), ( k ) -> new ArrayList<>() ).add( tariffGroup );
            }
        }

        return tariffGroupMap;
    }

	/**
	 * Получение списка тарифов (глобальных и персональных) договора.
	 * @return
	 * @throws BGException
	 */
	public List<ContractTariffEntry> getContractTariffEntryList()
		throws BGException
	{
		return contractTariffEntryList.get();
	}

	/**
	 * Отмена запланированного перехода с одного тарифа на другой.
	 * @param contractTariffId ID ContractTariff, который отменяем.
	 * @throws BGException
	 */
	public void cancel( int contractTariffId )
		throws BGException
	{
		ContractTariff contractTariff = contractTariffService.contractTariffGet( contractTariffId );
		if( contractTariff == null || contractTariff.getContractId() != getContractId() )
		{
			addErrorMessage( null, "Данный тарифный план не найден на договоре." );
			return;
		}

		if( contractTariff.getReplacedFromContractTariffId() <= 0 )
		{
			addErrorMessage( null, "Данный тарифный план нельзя отменить." );
			return;
		}

		ContractTariff oldContractTariff = contractTariffService.contractTariffGet( contractTariff.getReplacedFromContractTariffId() );
		if( oldContractTariff == null || oldContractTariff.getContractId() != getContractId() )
		{
			addErrorMessage( null, "Не найдена запись о предыдущем тарифе (contractTariffId=" + contractTariff.getReplacedFromContractTariffId() + ")." );
			return;
		}

		contractTariffService.contractTariffCancelByCustomer( getContractId(), contractTariff.getId() );

		populate();
	}

	/**
	 * Сейчас не используется - вместо него используется {@link #getTariffPlanTitle(int, String)}
	 * @param tariffPlanId
	 * @return
	 * @throws BGException
	 */
	public String getTariffPlanTitle( int tariffPlanId )
		throws BGException
	{
	    return getTariffPlanTitle(tariffPlanId,null);
	}
	    

	/**
	 * Возвращает название тарифа
	 * @param tariffPlanId ID тарифа
	 * @param title название персонального тарифа
	 * @return
	 * @throws BGException
	 */
	public String getTariffPlanTitle( int tariffPlanId, String title )
		throws BGException
	{
		TariffPlan tariffPlan = tariffPlanDirectory.get( tariffPlanId );
        if( tariffPlan == null ) // если тарифа нет в справочнике - похоже это персональный - возвращаем его title
		{
			return title;
		}

		return tariffPlan.isUseTitleInWeb() ? tariffPlan.getTitle() : tariffPlan.getTitleWeb();
	}
}
